import hashlib
import numpy as np
import tqdm
from elasticsearch import Elasticsearch
from elasticsearch.helpers import streaming_bulk
from gensim.models import KeyedVectors

es = Elasticsearch()


def insert_vocab(model):
    progress = tqdm.tqdm(unit='grams', total=len(model.wv.vocab))
    for ok, action in streaming_bulk(
            client=es, index='vocabulary', actions=create_docs(model),
    ):
        progress.update(1)
    print('Finished')


def create_docs(model):
    for word in model.wv.vocab:
        doc = {
            '_id': hashlib.md5(word.encode()).hexdigest(),
            'word': word,
            'vector': model.wv.word_vec(word, use_norm=True)
        }
        yield doc


def search_similar_terms(search_phrase):
    w2v_model = KeyedVectors.load('w2v_vocab.model')
    w2v_model.init_sims(replace=True)
    vector = None
    if len(search_phrase.split()) > 1:
        word_vecs = []
        for word in search_phrase.split():
            if word in w2v_model.wv.vocab:
                vec = w2v_model.wv.word_vec(word, use_norm=True)
                word_vecs.append(vec)
        vector = np.mean(word_vecs, axis=0).tolist()
    else:
        vector = w2v_model.wv.word_vec(search_phrase, use_norm=True).tolist()

    query = {
      "size": 5,
      "query": {
        "script_score": {
          "query": {
            "match_all": {}
          },
          "script": {
            "source": "cosineSimilarity(params.queryVector, 'vector')+1.0",
            "params": {
              "queryVector": vector
            }
          }
        }
      }
    }

    res = es.search(index="vocabulary", body=query)
    return res


def search_bm25_extra_grams(search_phrase, committee_type=None, item_type=None, date_range=None):
    res_vocab = search_similar_terms(search_phrase)
    similar_grams = [hit['_source']['word'] for hit in res_vocab['hits']['hits']]
    must_match = []
    if item_type:
        must_match.append({"match": {"item_type": item_type}},)
    if committee_type:
        must_match.append({"match": {"committee_type": committee_type}})
    if date_range:
        must_match.append({"range": {"date": {"gte": date_range[0],
                                              "lte": date_range[1]}}})
    must_match.append(
        {
            "nested": {  # to treat each paragraph as a separate document, specify nested path
                "path": "paragraphs",
                "query": {
                    "more_like_this": {
                        "fields": ["paragraphs.paragraph"],
                        "like": similar_grams,  # a list of search terms
                        "min_term_freq": 1,  # min amount of times for term to appear
                        "max_query_terms": 10  # range for terms to appear
                    }
                },
                "inner_hits": {
                    "size": 10,  # max relevant paragraphs
                    "highlight": {
                        "fields": {
                            "paragraphs": {}  # to return relevant paragraphs use highlight
                        }
                    }}}}
    )
    query = {
      "_source": ["date", "url", "committee_type", "item_type"],  # json fields to return. Delete the line to return all
      "size": 10,  # amount of docs to return
      "sort": [{
          "date": {"order": "desc"}}  # sort results by most recent first
      ],
      "query": {
        "bool": {
          "must": must_match}}}

    res = es.search(index="documents_test", body=query)
    return res
